/*****************************************************************************
 *   irq.c: Interrupt handler C file for NXP LPC23xx/24xx Family Microprocessors
 *
 *   Copyright(C) 2006, NXP Semiconductor
 *   All rights reserved.
 *
 *   History
 *   2006.07.13  ver 1.00    Prelimnary version, first Release
 *
 ******************************************************************************/ 

/* TODO mthomas - volatiles for vect_addr, vect_cntl? */

#include "config.h"
#include "inttypes.h"
#include "irq.h"

/* Initialize the interrupt controller */
/******************************************************************************
** Function name:		init_VIC
**
** Descriptions:		Initialize VIC interrupt controller.
** parameters:			None
** Returned value:		None
** 
******************************************************************************/
void init_VIC(void) 
{
    uint32_t i = 0;
    uint32_t *vect_addr, *vect_cntl;
   	
    /* initialize VIC*/
    VICIntEnClr = 0xffffffff;
    VICVectAddr = 0;
    VICIntSelect = 0;

    /* set all the vector and vector control register to 0 */
    for ( i = 0; i < VIC_SIZE; i++ )
    {
		vect_addr = (uint32_t *)(VIC_BASE_ADDR + VECT_ADDR_INDEX + i*4);
		vect_cntl = (uint32_t *)(VIC_BASE_ADDR + VECT_CNTL_INDEX + i*4);
		*vect_addr = 0x0;	
		*vect_cntl = 0xF;
    }
    return;
}

/******************************************************************************
** Function name:		install_irq
**
** Descriptions:		Install interrupt handler
** parameters:			Interrupt number, interrupt handler address, 
**						interrupt priority
** Returned value:		true or false, return false if IntNum is out of range
** 
******************************************************************************/
uint32_t install_irq( uint32_t IntNumber, void *HandlerAddr, uint32_t Priority )
{
    uint32_t *vect_addr;
    uint32_t *vect_cntl;
      
    VICIntEnClr = 1 << IntNumber;	/* Disable Interrupt */
    if ( IntNumber >= VIC_SIZE )
    {
		return ( FALSE );
    }
    else
    {
		/* find first un-assigned VIC address for the handler */
		vect_addr = (uint32_t *)(VIC_BASE_ADDR + VECT_ADDR_INDEX + IntNumber*4);
		vect_cntl = (uint32_t *)(VIC_BASE_ADDR + VECT_CNTL_INDEX + IntNumber*4);
		*vect_addr = (uint32_t)HandlerAddr;	/* set interrupt vector */
		*vect_cntl = Priority;
		VICIntEnable = 1 << IntNumber;	/* Enable Interrupt */
		return( TRUE );
    }
}

/******************************************************************************
                                  End Of File
******************************************************************************/

#ifndef IRQ_H_
#define IRQ_H_
/******************************************************************************
 *   irq.h:  Interrupt related Header file for NXP LPC23xx/24xx Family 
 *   Microprocessors
 *
 *   Copyright(C) 2006, NXP Semiconductor
 *   All rights reserved.
 *
 *   History
 *   2006.09.01  ver 1.00    Prelimnary version, first Release
 *
******************************************************************************/
#ifndef __IRQ_H 
#define __IRQ_H

#define I_Bit			0x80
#define F_Bit			0x40

#define SYS32Mode		0x1F
#define IRQ32Mode		0x12
#define FIQ32Mode		0x11

#define HIGHEST_PRIORITY	0x01
#define LOWEST_PRIORITY		0x0F

#define	WDT_INT			0
#define SWI_INT			1
#define ARM_CORE0_INT		2
#define	ARM_CORE1_INT		3
#define	TIMER0_INT		4
#define TIMER1_INT		5
#define UART0_INT		6
#define	UART1_INT		7
#define	PWM0_1_INT		8
#define I2C0_INT		9
#define SPI0_INT		10			/* SPI and SSP0 share VIC slot */
#define SSP0_INT		10
#define	SSP1_INT		11
#define	PLL_INT			12
#define RTC_INT			13
#define EINT0_INT		14
#define EINT1_INT		15
#define EINT2_INT		16
#define EINT3_INT		17
#define	ADC0_INT		18
#define I2C1_INT		19
#define BOD_INT			20
#define EMAC_INT		21
#define USB_INT			22
#define CAN_INT			23
#define MCI_INT			24
#define GPDMA_INT		25
#define TIMER2_INT		26
#define TIMER3_INT		27
#define UART2_INT		28
#define UART3_INT		29
#define I2C2_INT		30
#define I2S_INT			31

#define VIC_SIZE		32

#define VECT_ADDR_INDEX	0x100
#define VECT_CNTL_INDEX 0x200

/* Be aware that, from compiler to compiler, nested interrupt will have to
be handled differently. More details can be found in Philips LPC2000
family app-note AN10381 */
 

// mthomas
#ifdef __GNUC__
#define __irq
#define IENABLE 
#define IDISABLE
#else
/* unlike Keil CARM Compiler, in ARM's RealView compiler, don't save and 
restore registers into the stack in RVD as the compiler does that for you. 
See RVD ARM compiler Inline and embedded assemblers, "Rules for 
using __asm and asm keywords. */
static uint32_t sysreg;		/* used as LR register */
#define IENABLE __asm { MRS sysreg, SPSR; MSR CPSR_c, #SYS32Mode }
#define IDISABLE __asm { MSR CPSR_c, #(IRQ32Mode|I_Bit); MSR SPSR_cxsf, sysreg }
#endif

void init_VIC( void );
uint32_t install_irq( uint32_t IntNumber, void *HandlerAddr, uint32_t Priority );

#endif /* end __IRQ_H */

/******************************************************************************
**                            End Of File
******************************************************************************/

#endif /*IRQ_H_*/
 ...
 
// ========================================================================================================================
 Issue with LPC2378 IRQ interrupts
Chetan RP

Hi there,
I have 2 IRQ related ISR's in the program shown below.
ISR's 1 and 2 do the task of making an LED (connected to P2.0 and P2.7) on and off.
Iam seeing that when ISR1 is in execution(which is invoked by pressing a switch 
connected to P2.10 meant as external interrupt 0) and if ISR2 is invoked
(which is invoked by pressing a switch connected to P2.11 meant as external interrupt 1) 
during that time, after completion of ISR1, the control branches to ISR2. Could anyone 
please suggest a method to prevent this. 

#include <LPC23xx.H>

void wait(void){
  unsigned int i,j;

  for(i=0;i<=100;i++)
           for(j=0;j<=60000;j++);
}

/* External ISR 1  */
__irq void eint0_handler (void) {

  FIO2PIN = 0x00000001;
  wait();
  FIO2PIN = 0x00000000;
  EXTINT  = 0x00000001;
  VICVectAddr = 0;
}

/* External ISR 2 */
__irq void eint1_handler (void)  {

  FIO2PIN = 0x00000080;
  wait();
  FIO2PIN = 0x00000000;
  EXTINT  = 0x00000002;
  VICVectAddr = 0;
}

/* Initialize External Interrupts EINT 0/1 */
void init_eint (void) {
        EXTMODE      = 0x00000003;                           // Edge sensitive mode
        EXTPOLAR     = 0x00000000;                           // Active low, Falling edge

        VICIntSelect = 0x00000000;

        VICVectAddr14 = (unsigned long) eint0_handler;
        VICVectAddr15 = (unsigned long) eint1_handler;

        VICVectPriority14 = 0;
        VICVectPriority15 = 1;

        VICIntEnable = 0x0000C000;
}


int main (void) {

  PINSEL4 = 0x00500000;
  FIO2DIR = 0x000000FF;
  FIO2PIN = 0x00000000;

  init_eint();                                                            // Enable EINT0/1

   while (1){
           ;
   }
}

// -----------------------------------------------------------------------------------------------------------------------------------------------------------
LPC2368 ARM 7 TDMI MCB 2300
Samjith Wasim

I was able to configure and work with I2C interface in LPC2368 using 
Polling Method.There was no problem with the polling method.

Now i am forced to use I2C using Interrupt Routines for Real-Time Applications.
I configured the following registers for enabling ISR , when i tried simulating
the program , the VIC Address, VIC Enable Bit and VIC Priority are setting 
up properly.But the function where i used ISR is not invoking , i was not able 
to solve out where is the problem.Is there any more additional registers 
i should configure other than these ?

VICIntSelect = 0; // for selecting IRQ
VICIntEnable = (1<<19); //Enabling I2C1
VICVectCntl0 = 0x01; // High Priority
VICVectAddr0 = (unsigned long)I2C_IRQHandler;

__irq void I2C_IRQHandler()
{ ....
......

......
VICVectAddr = 0x00; //Acknowledge Interrupt
}

/* I am not getting the IRQ Function invoked at all.

Can any one help me out on this Issue as soon as possible ?

Regards,
Samjith */

// -------------------------------------------------------------------------------------------------------------------------------------------------------

LPC2368 ARM 7 TDMI MCB 2300
Samjith Wasim

VICIntSelect = 0; // for selecting IRQ
VICIntEnable = (1<<19); //Enabling I2C1
VICVectCntl19 = 0x01; // High Priority
VICVectAddr19 = (unsigned long)I2C_IRQHandler;

__irq void I2C_IRQHandler()
{ ....
......

......
VICVectAddr = 0x00; //Acknowledge Interrupt
}

/* Is this Configuration Correct ?? But its still not Working Sir.

Regards,
Samjith */

// -------------------------------------------------------------------------------------------------------------------------------------------------------
LPC2378 crash after interrupt
Demis Biscaro

Yesterday I've opened a thread about a system crash after the execution of an IRQ routine.
After many tests I've isolated the problem and symplified the context in which it happens.

Now the problem I see is the following.

I've only an IRQ routine linked to CPU timer 0. It should be invoked every 1 ms. 
However after the first call the CPU switches to DATA ABORT mode.

I configure LPC2378 (rev. B) with 72 MHz clock and MAM fully enabled 
(the problem is present olaso with MAM partially enabled).
The source code is the following

/*************************************************
** INITIALISING FUNCTION
**************************************************/
void ini_base_timer(void){
    PCONP |= (1 << 1); /* power up timer 0 */
    T0TCR = 0;       /* disable timer 0 */
    PCLKSEL0 |= 0x00000004;      /* timer 0 clock = PCLOCK */
    T0IR = 0x11;        /* reset irq flag */
    T0CTCR = 0;
    T0PR = 8;   /* prescaler */
    T0MR0 = 7999;       /* match register 0 */

/* configuration vectored interrupt */
    VICIntEnClr |= (1 << 4);  /* irq disabling */
    VICIntSelect &= ~(1 << 4);/* IRQ interrupt */
    VICVectPriority4 = 8;          /* priority */
    VICVectAddr4 = (UINT32)&irq_base_timer;         /* ISR */
    VICIntEnable |= (1 << 4);/* irq enabling*/

    T0MCR |= 0x03;       /* irq enabling and TC reset on match register 0 value */
    T0TCR = 0x01;       /* start timer */
}   /* end ini_base_timer */


/*************************************************
** IRQ FUNCTION (ISR)
** global variables:
** long sd[32];                     // global array
** unsigned char buffer_rx_fw[512]; // global array
**************************************************/
void __IRQ__ irq_base_timer(void){
        UINT32 cont;

         FIO4SET = 0x01; /* set GPOUTPUT 4.0 */
         T0IR = 1;      /* irq flag reset */

        for(cont = 0; cont < 2; cont++){
            if(sd[31] == cont){
              sd[30]++;
              if(sd[30] == buffer_rx_fw[cont])
                 sd[28]++;
           } /* end if */
        } /* end for */


         FIO4CLR = 0x01; /* clear GPOUTPUT 4.0 */

         VICVectAddr = 0;    /* dummy write */
}   /* end irq_base_timer */


In this condition the problem is deterministic in the sense that every time after 
the first execution of irq function, CPU switches to DATA ABORT mode (the system crashes!).
If I execute code step-by-step in assembly mode I see that when micro exits from 
irq function the LR register contains the address of an instruction that causes DATA ABORT exception.
I don't know how a simple IRQ routine can generate such terrible problem!

I'm using GCC compiler (ver. 4.0.3) with optimization level 1.
If I put optimization level to 0 (zero) the system doesn't crash, however the assembly 
code produced is so redundant to degradate system performance too much!

Have you any suggestions?

Thank you,

Demis

// ======================================================================================================
ARM LPC2378: data/prefetch abort after interrupt
Demis Biscaro

I have a problem with ARM LPC2378 revision B, configured with CPU clock 
frequency at 72 MHZ.
I have the following configuration:
- an IRQ interrupt function linked to CPU timer 0 (vectored interrupt). 
It is invoked with a period of 1 ms
- an FIQ interrupt function linked to the half-empty flag of RX queue 
of MCI bus controller (the bus used to communicate with SD-MMC cards).

During the reading of a file from an SD card CPU switches to 
"Data Abort" or "Prefetch Abort". While CPU is in this error condition 
I very often see that IRQ stack pointer is corrupted, i.e. it points 
to an invalid address. However this doesn't happens every time, but only very often.

I've controlled that stack dimension of each mode is well sized.

Inside the IRQ function some other functions are invoked. If I rewrite 
the code of these inner functions as "inline code" inside the IRQ function 
the system doesn't crash!
It seems the problem to be the call of a function inside the IRQ function 
but I'm not sure of this conclusion.

The problem happens both with MAM fully enabled and MAM partially enabled.

Any suggestions?

Thanks and have a nice day!

Demis

/***********************************************
** This is the IRQ function invoked every 1 ms
************************************************/
void __IRQ__ irq_base_timer(void){
   T0IR = 1; /* reset interrupt flag */

   test_fun();


   VICVectAddr = 0; /* dummy write */
}

/***********************************************
** This is the function invoked inside
** irq_base_timer
**
** long sd[32];            // global array
** char buffer_rx_fw[512]; // global array
**
************************************************/
void test_fun(void){
   unsigned long cont;
/* dummy cycle */
   for(cont = 0; cont < 500; cont++){
      if(sd[31] == cont){
         sd[30]++;
         if(sd[29] == buffer_rx_fw[cont])
            sd[28]++;
      }
   }
}


/***********************************************
** This is the FIQ function used to deploy the
** MCI RX FIFO
**
** long* p_sd_fifo; // global pointer to FIFO
** char* p_sd_data_rx; // global pointer to RAM buffer
** long sd_long_count; // global long counter
**
************************************************/
void __FIQ__ irq_sd_rx(void){
   unsigned char i;
   union{unsigned long L; unsigned int W[2]; unsigned char B[4];} tmp;

   for(i = 0; i < 8; i++){
      tmp.L = *p_sd_fifo++;
      *p_sd_data_rx++ = tmp.B[0];
      *p_sd_data_rx++ = tmp.B[1];
      *p_sd_data_rx++ = tmp.B[2];
      *p_sd_data_rx++ = tmp.B[3];
   }    /* end for */
   if(sd_long_count & 1)
      p_sd_fifo = (long*)0xE008C080; // FIFO start address
   sd_long_count--;
}   /* end irq_sd_rx */

// =========================================================================================================

Wlasnie rozgryzam tego arma, popelnilem taki kod opierajac sie na zrodlach z netu:


#include "LPC23xx.H"
#define LEDPIN  18   //D7 on KT-M-LPC2378 Kristech

volatile int i;

void t0isr(void) __attribute__ ((interrupt("IRQ")));
void t0isr(void)
{
  if (i==1) FIO1CLR = (1<<LEDPIN);
  if (i==0) FIO1SET = (1<<LEDPIN);
  if (i==1) i=0;
  if (i==0) i=1;
 // T0IR = 0x01;
 // VICVectAddr = 0x00000000;

}



int main (void)
{

   i=0;
   FIO1DIR |= (1<<LEDPIN);
   FIO1CLR = (1<<LEDPIN);
      T0TCR = 0x02;                           /* Reset timer */
      T0IR = 0xff;                           /* Reset all interrupts   */
      T0MCR = 0x03;                           /* Reset TC on match */
      T0MR0 = 0x125100;                     /* Set match register (1mSec = 15.000-1 counts) */
      VICIntSelect = 0x00;
      VICVectAddr4 = (unsigned)t0isr;
      VICVectPriority4 = 0x0000000010;
         VICIntEnable = 0x10;                  /* Enable Timer0 interrupt */
      T0TCR = 0x01;                           /* Start timer */


      while(1)
      {                              /* Loop forever */

      }



   return 0;
}
// -------------------------------------------------------------------------------------------

To Ci powinno ruszy, napisaem kiedys pod Keila uVision3 V3.63. 
Dziala z Makieta Olimex'u LPC-2378-STK-A: 

// (LPC2378) mandaryn 2009

#include <lpc23xx.h>

int flaga=0;                                                // 0 -> dioda OFF, 1 -> dioda ON

__irq void przerwanie (void)
{
   if(flaga==0)                                             // Gdy dioda wylaczona
   {
      FIO2SET|=0x00000001;                                  // Wlaczenie diody
      flaga=1;                                              // flaga -> dioda ON
   }
   else                                                     // Gdy dioda wlaczona
   {
      FIO2CLR|=0x00000001;                                  // Wylaczenie diody
      flaga=0;                                              // flaga -> dioda OFF
   }
    T0IR |= 8;                                              // wyzerowanie flagi przerwania
    VICVectAddr = 0;                                        // wyczyszczenie VICAddress
}
   
int main(void)
{
    T0MR3 = 15*1000*140-1;                       // wartosc: (czekaj 140 ms)
    T0MCR = 0x00000600;                          // licznik gdy osiagnie pozadana wartosc: przerwanie + reset
   
    FIO2DIR|=0x00000001;                         // Ustawienie portu Fast2.0 (dioda podlaczona do EXT2, pin 37)

    VICVectAddr4=(unsigned long)przerwanie;      // Wpisanie adresu przerwania od Timar0   (zawsze kanal 4)
    VICVectPriority4 = 0x0000000001;             // Ustawienie priorytetu dla przerwania od Timer0 (kanalu4)
    VICIntEnable=0x00000010;                     // Wlaczenie przerwania

    T0TCR = 0x00000001;                          // wlaczenie licznika
      
    while(1);                                    // nieskonczona petla
}

// ------------------------------------------------------------------------------------

Witam, mam napisany program na LPC2378 w ktrym uywam timera T0 ktry generuje przerwania 
a raczej generowa. 


  T0PR = PR_1MS; // 1ms
  T0TCR = RESET_PR_CNT; // reset counter and prescaler
  T0MCR = INTERRUPT_STOP_ON_MATCH;
  T0MR2 = 1000;    // reset co 1s

  VICVECTADDR4 = (unsigned)( T0isr);
  VICVECTPRIORITY4 = 0x2;
  VICINTENABLE = 0x10;
 
  T0TCR = T0_ON;

obsuga przerwania:

__irq __arm void T0isr(void)
{
  Meas = true;
  T0TC = 0;
  T0IR |= 4; // clear interrupt flag
  T0TCR = T0_ON;
  VICADDRESS = 0x00000000;
} 

// ----------------------------

msr   CPSR_c,#SVC_MODE|I_BIT|F_BIT       // ... to supervisor mode 

// =========================================================

#include "LPC23xx.H"
#define LEDPIN  18   //D7 on KT-M-LPC2378 Kristech

volatile int i;

void t0isr(void) __attribute__ ((interrupt("IRQ")));
void t0isr(void)
{
  if (i==1) FIO1CLR = (1<<LEDPIN);
  if (i==0) FIO1SET = (1<<LEDPIN);
  if (i==1) i=0;
  if (i==0) i=1;
 // T0IR = 0x01;
 // VICVectAddr = 0x00000000;

}



int main (void)
{

   i=0;
   FIO1DIR |= (1<<LEDPIN);
   FIO1CLR = (1<<LEDPIN);
      T0TCR = 0x02;                           /* Reset timer */
      T0IR = 0xff;                           /* Reset all interrupts   */
      T0MCR = 0x03;                           /* Reset TC on match */
      T0MR0 = 0x125100;                     /* Set match register (1mSec = 15.000-1 counts) */
      VICIntSelect = 0x00;
      VICVectAddr4 = (unsigned)t0isr;
      VICVectPriority4 = 0x0000000010;
         VICIntEnable = 0x10;                  /* Enable Timer0 interrupt */
      T0TCR = 0x01;                           /* Start timer */


      while(1)
      {                              /* Loop forever */

      }



   return 0;
} 

// ------------------------------------------------------------------------------------------------
// (LPC2378) mandaryn 2009

#include <lpc23xx.h>

int flaga=0;                                                // 0 -> dioda OFF, 1 -> dioda ON

__irq void przerwanie (void)
{
   if(flaga==0)                                             // Gdy dioda wylaczona
   {
      FIO2SET|=0x00000001;                                  // Wlaczenie diody
      flaga=1;                                              // flaga -> dioda ON
   }
   else                                                     // Gdy dioda wlaczona
   {
      FIO2CLR|=0x00000001;                                  // Wylaczenie diody
      flaga=0;                                              // flaga -> dioda OFF
   }
    T0IR |= 8;                                              // wyzerowanie flagi przerwania
    VICVectAddr = 0;                                        // wyczyszczenie VICAddress
}
   
int main(void)
{
    T0MR3 = 15*1000*140-1;                       // wartosc: (czekaj 140 ms)
    T0MCR = 0x00000600;                          // licznik gdy osiagnie pozadana wartosc: przerwanie + reset
   
    FIO2DIR|=0x00000001;                         // Ustawienie portu Fast2.0 (dioda podlaczona do EXT2, pin 37)

    VICVectAddr4=(unsigned long)przerwanie;      // Wpisanie adresu przerwania od Timar0   (zawsze kanal 4)
    VICVectPriority4 = 0x0000000001;             // Ustawienie priorytetu dla przerwania od Timer0 (kanalu4)
    VICIntEnable=0x00000010;                     // Wlaczenie przerwania

    T0TCR = 0x00000001;                          // wlaczenie licznika
      
    while(1);                                    // nieskonczona petla
}

// ========================================================================================